#pragma once
/**
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0.
 */

#define AWS_CRT_CPP_VERSION "0.37.2"
#define AWS_CRT_CPP_VERSION_MAJOR 0
#define AWS_CRT_CPP_VERSION_MINOR 37
#define AWS_CRT_CPP_VERSION_PATCH 2
#define AWS_CRT_CPP_GIT_HASH ""
